#!/bin/bash
#
#	fcalltable.sh
#
#		Shows the per-syscall calls to the schedule() function
#		getting its data from the /proc/PID/sched file. In
#		order to get the system calls' names, it does need to
#		read arch/x86/ / header file.
#
#	2012 by Toni Castillo Girona
#	<toni.castillo@fa.upc.edu>
#
#
#set -x

LINUXSOURCE="/usr/src/linux-source-2.6.32"					# Kernel source tree
SYSCALLNAMES="arch/x86/include/asm/unistd_64.h"				# Syscall defines, C-header file
NSYSCALL_MAX=299											# nsyscall_max + 1

#----------------------------------------------------------------------------------
#	get_syscall_name
#		Return the name for a particular system call id.
#----------------------------------------------------------------------------------
get_syscall_name () {
	if [ -r $LINUXSOURCE/$SYSCALLNAMES ]; then
		# Get all the syscall names, ordered by index, so the first one is index=0,
		# the next one index=1, and so on
		index=`expr $1 + 1`			# first 0 + 1 = 1 ...
		sname=`cat $LINUXSOURCE/$SYSCALLNAMES|grep -E "^#define __NR_"|tr  \
									'\t' ' '|cut -d" " -f2|head -$index|tail -1`
		echo "${sname:5}"
	else
		echo ""		# Empty string
		return 1	# error
	fi
}

#
# We need the pid to get the call table for
#
if [ $#  -lt 1 ]; then
	echo "Usage: `basename $0` PID [-s]\n"
	exit 1
fi

# Be careful, the pid must exist:
if [ -r /proc/$1/sched ]; then

	clear
	DATA=""

	# print task's pid and name:
	echo -ne "Task: $1\nCommand: `cat /proc/$1/cmdline`\nTotal calls: \
			 `cat /proc/$1/sched|grep nr_vol_switches_due_to_syscalls |tr -d ' ' \
			|cut -d":" -f2`\n"
	echo ""

	#Header:
	test "$2" != "-s" &&  printf "%-25s %15s \n"	"System Call"	"Calls to schedule()"

	# Get table's values:
	table=`cat /proc/$1/sched|grep -E "TCG.table."|tr -d ' '| \
			cut -d":" -f2|tr '\n' ' '`
	itbl=0
	for row in $table;
	do
		test "$2" != "-s" && printf	'%-25s %15d \n'  "`get_syscall_name $itbl`" "$row"
		# Get rows with >0 
		if [ $row -ge 1 ]; then
			rs=`printf '%-25s %15d \n' "\`get_syscall_name $itbl\`" "$row"`
			DATA="$DATA\n$rs"		
		fi
		itbl=`expr $itbl + 1`
	done
	# Print the summary:
	if [ ! -z "$DATA" ]; then 
		echo "-------------------------------------------"
		echo "Summary"
		echo "-------------------------------------------"
		printf "%-25s %15s"    "System Call"   "Calls to schedule()"
		echo -ne "$DATA\n"
	fi
	# Print the calls actually issuing calls to the schedule() function:
else
	echo "Error: /proc/$1/sched does not exist."
	exit 1
fi
