/* getdelays.c
 *
 * Utility to get per-pid and per-tgid delay accounting statistics
 * Also illustrates usage of the taskstats interface
 *
 * Copyright (C) Shailabh Nagar, IBM Corp. 2005
 * Copyright (C) Balbir Singh, IBM Corp. 2006
 * Copyright (c) Jay Lan, SGI. 2006
 *
 * Compile with
 *	gcc -I/usr/src/linux/include getdelays.c -o getdelays
 * 2012 Modified by Toni Castillo Girona <toni.castillo@fa.upc.edu>
 *
 *	Based on "getdelays.c", this program communicates with the GNU/Linux Kernel using NETLINNK
 *  sockets interface, in order to obtain "taskstats" fields.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <poll.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <signal.h>

#include <linux/genetlink.h>
#include <linux/taskstats.h>
#include <linux/cgroupstats.h>

/*
 * Generic macros for dealing with netlink sockets. Might be duplicated
 * elsewhere. It is recommended that commercial grade applications use
 * libnl or libnetlink and use the interfaces provided by the library
 */
#define GENLMSG_DATA(glh)	((void *)(NLMSG_DATA(glh) + GENL_HDRLEN))
#define GENLMSG_PAYLOAD(glh)	(NLMSG_PAYLOAD(glh, 0) - GENL_HDRLEN)
#define NLA_DATA(na)		((void *)((char*)(na) + NLA_HDRLEN))
#define NLA_PAYLOAD(len)	(len - NLA_HDRLEN)

#define	MIN_VERSION	8		// Minimal Taskstats struct version

#define err(code, fmt, arg...)			\
	do {					\
		fprintf(stderr, fmt, ##arg);	\
		exit(code);			\
	} while (0)

char name[100];
int dbg;						// 1= show messages
int involuntary = 0;

#define PRINTF(fmt, arg...) {			\
	    if (dbg) {				\
		printf(fmt, ##arg);		\
	    }					\
	}

/* Maximum size of response requested or message sent */
#define MAX_MSG_SIZE	1024

/* TCG: default delay time for gathering messages in seconds */
#define DEFAULT_DELAY	5

/* Maximum number of cpus expected to be specified in a cpumask */
#define MAX_CPUS    32
char cpumask[100+6*MAX_CPUS];

struct msgtemplate {
	struct nlmsghdr n;
	struct genlmsghdr g;
	char buf[MAX_MSG_SIZE];
};

static void usage(void)
{
	fprintf(stderr, "getcsw -p pid [-i] [-l] [-d secs] [-m cpumask] [-v]\n");
	fprintf(stderr, "  -i: show involuntary context switches extended counters\n");
	fprintf(stderr, "  -l: listen forever\n");
	fprintf(stderr, "  -d: delay in second for -l flag\n");
	fprintf(stderr, "  -m: sets the cpu(s) where we are listening to tasks exiting.\n");
	fprintf(stderr, "  -v: debug on\n");
}

/*
 * Create a raw netlink socket and bind
 */
static int create_nl_socket(int protocol)
{
	int fd;
	struct sockaddr_nl local;

	fd = socket(AF_NETLINK, SOCK_RAW, protocol);
	if (fd < 0)
		return -1;

	memset(&local, 0, sizeof(local));
	local.nl_family = AF_NETLINK;

	if (bind(fd, (struct sockaddr *) &local, sizeof(local)) < 0)
		goto error;

	return fd;
error:
	close(fd);
	return -1;
}


static int send_cmd(int sd, __u16 nlmsg_type, __u32 nlmsg_pid,
	     __u8 genl_cmd, __u16 nla_type,
	     void *nla_data, int nla_len)
{
	struct nlattr *na;
	struct sockaddr_nl nladdr;
	int r, buflen;
	char *buf;

	struct msgtemplate msg;

	msg.n.nlmsg_len = NLMSG_LENGTH(GENL_HDRLEN);
	msg.n.nlmsg_type = nlmsg_type;
	msg.n.nlmsg_flags = NLM_F_REQUEST;
	msg.n.nlmsg_seq = 0;
	msg.n.nlmsg_pid = nlmsg_pid;
	msg.g.cmd = genl_cmd;
	msg.g.version = 0x1;
	na = (struct nlattr *) GENLMSG_DATA(&msg);
	na->nla_type = nla_type;
	na->nla_len = nla_len + 1 + NLA_HDRLEN;
	memcpy(NLA_DATA(na), nla_data, nla_len);
	msg.n.nlmsg_len += NLMSG_ALIGN(na->nla_len);

	buf = (char *) &msg;
	buflen = msg.n.nlmsg_len ;
	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;
	while ((r = sendto(sd, buf, buflen, 0, (struct sockaddr *) &nladdr,
			   sizeof(nladdr))) < buflen) {
		if (r > 0) {
			buf += r;
			buflen -= r;
		} else if (errno != EAGAIN)
			return -1;
	}
	return 0;
}


/*
 * Probe the controller in genetlink to find the family id
 * for the TASKSTATS family
 */
static int get_family_id(int sd)
{
	struct {
		struct nlmsghdr n;
		struct genlmsghdr g;
		char buf[256];
	} ans;

	int id = 0, rc;
	struct nlattr *na;
	int rep_len;

	strcpy(name, TASKSTATS_GENL_NAME);
	rc = send_cmd(sd, GENL_ID_CTRL, getpid(), CTRL_CMD_GETFAMILY,
			CTRL_ATTR_FAMILY_NAME, (void *)name,
			strlen(TASKSTATS_GENL_NAME)+1);

	rep_len = recv(sd, &ans, sizeof(ans), 0);
	if (ans.n.nlmsg_type == NLMSG_ERROR ||
	    (rep_len < 0) || !NLMSG_OK((&ans.n), rep_len))
		return 0;

	na = (struct nlattr *) GENLMSG_DATA(&ans);
	na = (struct nlattr *) ((char *) na + NLA_ALIGN(na->nla_len));
	if (na->nla_type == CTRL_ATTR_FAMILY_ID) {
		id = *(__u16 *) NLA_DATA(na);
	}
	return id;
}

/*
	TCG: this function has been modified so as to print some
		 another fields, by extending taskstats structure
		 in the Kernel
*/
static void task_context_switch_counts(struct taskstats *t)
{

	/* TCG: get voluntary or non-voluntary context switch due to exit: */
	char vol_exit='-';									// By default, "-"
	if(t->version>=MIN_VERSION){
		if(t->nvcsw_ext[1]>0)vol_exit='y';				// We have the counter set to 1!
		else
			vol_exit = (t->nivcsw_ext[1]>0)?'n':'-';	// Non-vol or task's still running!
	}

	// Because we consider vb to be nvvsw-nvcsw_ext[0], when syscalls is huge, maybe
	// nvcsw_ext[0] >> nvcsw. That means a lot of context switches when returning from a
	// syscall. So, number of Voluntary Blocking are really less than these ones. A negative
	// number will show its relation. 
	signed long vb=t->nvcsw-t->nvcsw_ext[0];

	// TCG; print out the task's stats
	if(involuntary){
		printf("%4d     %13llu%15llu%15llu%15llu\n",
		   (pid_t)t->ac_pid , 				// TCG: show task's pid
	       (unsigned long long)t->nvcsw, (unsigned long long)t->nivcsw,
		   (t->version>=MIN_VERSION)?t->nivcsw_ext[0]:0,		// TCG: print number of cs due to try_to_wake_up
		   (t->version>=MIN_VERSION)?t->nivcsw_ext[1]:0		// TCG: print number of cs due to IRQS
		);
	}else{
		printf("%4d     %13llu%15llu%15llu%15llu%15lld%15c\n",
		   (pid_t)t->ac_pid , 				// TCG: show task's pid
	       (unsigned long long)t->nvcsw, (unsigned long long)t->nivcsw,
			/* Get dummy if do we have the right version ... */
		   (t->version>=MIN_VERSION)?t->nsyscalls:0,		// TCG: print number of issued syscalls
		   (t->version>=MIN_VERSION)?t->nvcsw_ext[0]:0,		// TCG: print number of cs due to issued syscalls
		   /* Compute this value: let's imagine voluntary = nvcsw - (nvcsw_ext[0] +1) */
		   (t->version>=MIN_VERSION)?vb:0,
		   vol_exit											// TCG: print vol or non-vol cs due to exit
		);
	}
}

int main(int argc, char *argv[])
{
	int c, rc, rep_len, aggr_len, len2;
	int cmd_type = TASKSTATS_CMD_ATTR_UNSPEC;
	__u16 id;
	__u32 mypid;

	struct nlattr *na;
	int nl_sd = -1;
	int len = 0;
	pid_t tid = 0;
	pid_t rtid = 0;

	int fd = 0,
		count = 0,
		maskset = 0,
		loop = 0,
		delay=DEFAULT_DELAY;		// TCG: a delay for -l

	struct msgtemplate msg;

	while (1) {
		c = getopt(argc, argv, "p:vld:m:i");
		if (c < 0)
			break;

		/* TCG: now, we are gonna print ALWAYS the Context Switches Accounts */
		switch (c) {
		case 'p':									// Select desired task per pid
			tid = atoi(optarg);
			if (!tid)
				err(1, "Invalid pid\n");
			cmd_type = TASKSTATS_CMD_ATTR_PID;
			break;
		case 'd':									// Set the delay
			delay = atoi(optarg);
			if(!delay)
				err(1,"Invalid expressed delay in seconds\n");
			break;
		case 'v':									// Debug on|off
			dbg = 1;
			break;
		case 'l':									// Iterate forever
			loop = 1;
			break;
        case 'm':									// Set a cpu-mask to get stats from exiting processes
            strncpy(cpumask, optarg, sizeof(cpumask));
            maskset = 1;
            printf("cpumask %s maskset %d\n", cpumask, maskset);
			/* We are gonna set the loop */
			loop = 1;
            break;
		case 'i':
			involuntary = 1;
			break;
		default:
			usage();
			exit(-1);
		}
	}

	/* TCG If we don't have a valid pid, exit: */
	if(!tid){ usage(); exit(-1); }

	// Create the NetLink socket and bind it:
	if ((nl_sd = create_nl_socket(NETLINK_GENERIC)) < 0)
		err(1, "error creating Netlink socket\n");

	// Get my own PID and choose the right TASKSTATS family:
	mypid = getpid();
	id = get_family_id(nl_sd);
	if (!id) {
		fprintf(stderr, "Error getting family id, errno %d\n", errno);
		goto err;
	}PRINTF("family id %d\n", id);

	// Set the cpumask message:
	if (maskset) {
       rc = send_cmd(nl_sd, id, mypid, TASKSTATS_CMD_GET,
                  TASKSTATS_CMD_ATTR_REGISTER_CPUMASK,
                  &cpumask, strlen(cpumask) + 1);
       PRINTF("Sent register cpumask, retval %d\n", rc);
       if (rc < 0) {
            fprintf(stderr, "error sending register cpumask\n");
            goto err;
        }
    }

	// TCG: show the looping-delay time in secs:
	if(loop)PRINTF("Delay for looping: %d\n", delay);

	if(involuntary){
		printf("Task   %15s%15s%15s%15s\n" ,
	       "voluntary", "nonvoluntary", "wake_up", "IRQS"
		);
	}else{
		printf("Task   %15s%15s%15s%15s%15s%15s\n" ,
	       "voluntary", "nonvoluntary", "syscalls", "ret.syscalls", "vol.sched" , "vol.exit"
		);
	}

	// Iterate allways or run it just once:
	do {
		
		int i;

		/* Send the command to get our per-task stats : */
		/* but only if we are not listening to cpumask  messages ... */
		if (tid && !maskset) {
			rc = send_cmd(nl_sd, id, mypid, TASKSTATS_CMD_GET,
				      cmd_type, &tid, sizeof(__u32));
			PRINTF("Sent pid/tgid, retval %d\n", rc);
			if (rc < 0) {
				fprintf(stderr, "error sending tid/tgid cmd\n");
				goto err;
			}
		}

		// Get kernel's response:
		rep_len = recv(nl_sd, &msg, sizeof(msg), 0);
		PRINTF("received %d bytes\n", rep_len);

		// A reply error; we can just continue
		if (rep_len < 0) {
			fprintf(stderr, "nonfatal reply error: errno %d\n",
				errno);
			continue;
		}

		// Is it a NETLINK error message?
		if (msg.n.nlmsg_type == NLMSG_ERROR ||
		    !NLMSG_OK((&msg.n), rep_len)) {
			struct nlmsgerr *err = NLMSG_DATA(&msg);
			fprintf(stderr, "fatal reply error,  errno %d\n",
				err->error);
			goto err;
		}

		// Show some stats for this received message:
		PRINTF("nlmsghdr size=%zu, nlmsg_len=%d, rep_len=%d\n",
		       sizeof(struct nlmsghdr), msg.n.nlmsg_len, rep_len);
		rep_len = GENLMSG_PAYLOAD(&msg.n);

		na = (struct nlattr *) GENLMSG_DATA(&msg);		// Get attribute header
		len = 0; i = 0;

		while (len < rep_len) {
			len += NLA_ALIGN(na->nla_len);
			PRINTF("nla_type=%d\n" , na->nla_type);
			switch (na->nla_type) {
			case TASKSTATS_TYPE_AGGR_PID:				// We do have stats per -task:
				aggr_len = NLA_PAYLOAD(na->nla_len);
				len2 = 0;
				/* For nested attributes, na follows */
				na = (struct nlattr *) NLA_DATA(na);	// Get Attribute data
				while (len2 < aggr_len) {
					switch (na->nla_type) {
					case TASKSTATS_TYPE_PID:			// This attribute has a payload of PID.
						rtid = *(int *) NLA_DATA(na);
						break;
					case TASKSTATS_TYPE_TGID:
						rtid = *(int *) NLA_DATA(na);
						break;
					case TASKSTATS_TYPE_STATS:				// Taskstats struct is in this payload:
						count++;
						/* TCG: only print this data if it is coming from our PID! */
						if(tid && ((struct taskstats *) NLA_DATA(na))->ac_pid==tid){
							task_context_switch_counts((struct taskstats *) NLA_DATA(na));
							/* TCG: if we do have cpumask, we can get out of here right away: */
							if(maskset)loop=0;
						}else {
							/* Another task has exited, don't pay attention to it and resumes :*/
							if(maskset)goto resume;
						}
						if (!loop)							// Are we looping?
							goto err;
						break;
					default:								// Unknown nested attribute
						fprintf(stderr, "Unknown nested"
							" nla_type %d\n",
							na->nla_type);
						break;
					}
/* TCG Added to resume from a cpumkask condition: */
resume:
					len2 += NLA_ALIGN(na->nla_len);
					na = (struct nlattr *) ((char *) na + len2);
				}
				break;
			default:									// Unknown data
				fprintf(stderr, "Unknown nla_type %d\n",
					na->nla_type);
				break;
			}
			na = (struct nlattr *) (GENLMSG_DATA(&msg) + len);
		}

		/* TCG: white delay seconds : */
		sleep(delay);

	} while (loop);

/* Close file handles - socket - */
err:
	 /* TCG: Clean up the cpumask, though the kernel does that by itself after a while ;-) :*/
	 if (maskset) {
        rc = send_cmd(nl_sd, id, mypid, TASKSTATS_CMD_GET,
                  TASKSTATS_CMD_ATTR_DEREGISTER_CPUMASK,
                  &cpumask, strlen(cpumask) + 1);
        printf("Sent deregister mask, retval %d\n", rc);
        if (rc < 0)
            err(rc, "error sending deregister cpumask\n");
     }

	/* TCG: close the socket: */
	close(nl_sd);
	if (fd) close(fd);
	return 0;
}
